@@ -41,11 +41,13 @@ enum config_section {
 	GLOBAL_SECTION,
 	PORT_SECTION,
 	DEVICE_SECTION,
+	EXT_SECTION,
 	UNKNOWN_SECTION,
 };
 
 enum config_type {
 	CFG_TYPE_INT,
+	CFG_TYPE_U64,
 	CFG_TYPE_DOUBLE,
 	CFG_TYPE_ENUM,
 	CFG_TYPE_STRING,
@@ -60,16 +62,18 @@ typedef union {
 	int i;
 	double d;
 	char *s;
+	uint64_t u64;
 } any_t;
 
 
 #define CONFIG_LABEL_SIZE 32
 
-#define CFG_ITEM_STATIC (1 << 0) /* statically allocated, not to be freed */
-#define CFG_ITEM_LOCKED (1 << 1) /* command line value, may not be changed */
-#define CFG_ITEM_PORT   (1 << 2) /* item may appear in port sections */
-#define CFG_ITEM_DYNSTR (1 << 4) /* string value dynamically allocated */
-#define CFG_ITEM_DEVICE (1 << 8) /* item may appear in device sections */
+#define CFG_ITEM_STATIC	(1 << 0)  /* statically allocated, not to be freed */
+#define CFG_ITEM_LOCKED	(1 << 1)  /* command line value, may not be changed */
+#define CFG_ITEM_PORT	(1 << 2)  /* item may appear in port sections */
+#define CFG_ITEM_DYNSTR	(1 << 4)  /* string value dynamically allocated */
+#define CFG_ITEM_DEVICE	(1 << 8)  /* item may appear in device sections */
+#define CFG_ITEM_EXT	(1 << 16) /* item may appear in external sections */
 
 struct config_item {
 	char label[CONFIG_LABEL_SIZE];
@@ -85,6 +89,7 @@ struct config_item {
 
 #define PORT_TO_FLAG(_port) _port == PORT_SECTION ? CFG_ITEM_PORT : \
 	_port == DEVICE_SECTION ? CFG_ITEM_DEVICE : \
+	_port == EXT_SECTION ? CFG_ITEM_EXT : \
 	CFG_ITEM_STATIC
 
 #define CONFIG_ITEM_DBL(_label, _port, _default, _min, _max) {	\
@@ -110,6 +115,14 @@ struct config_item {
 	.min.i	= _min,					\
 	.max.i	= _max,					\
 }
+#define CONFIG_ITEM_U64(_label, _port, _default, _min, _max) {	\
+	.label	= _label,				\
+	.type	= CFG_TYPE_U64,			\
+	.flags	= PORT_TO_FLAG(_port),			\
+	.val.u64 = _default,				\
+	.min.u64 = _min,				\
+	.max.u64 = _max,				\
+}
 #define CONFIG_ITEM_STRING(_label, _port, _default) {	\
 	.label	= _label,				\
 	.type	= CFG_TYPE_STRING,			\
@@ -144,9 +157,18 @@ struct config_item {
 #define DEV_ITEM_INT(label, _default, min, max) \
 	CONFIG_ITEM_INT(label, DEVICE_SECTION, _default, min, max)
 
+#define DEV_ITEM_U64(label, _default, min, max) \
+	CONFIG_ITEM_U64(label, DEVICE_SECTION, _default, min, max)
+
 #define DEV_ITEM_STR(label, _default) \
 	CONFIG_ITEM_STRING(label, DEVICE_SECTION, _default)
 
+#define EXT_ITEM_INT(label, _default, min, max) \
+	CONFIG_ITEM_INT(label, EXT_SECTION, _default, min, max)
+
+#define EXT_ITEM_STR(label, _default) \
+	CONFIG_ITEM_STRING(label, EXT_SECTION, _default)
+
 struct config_item config_tab_synce[] = {
 	GLOB_ITEM_INT("logging_level", LOG_INFO, PRINT_LEVEL_MIN, PRINT_LEVEL_MAX),
 	GLOB_ITEM_STR("message_tag", NULL),
@@ -154,8 +176,6 @@ struct config_item config_tab_synce[] = {
 	GLOB_ITEM_STR("userDescription", ""),
 	GLOB_ITEM_INT("verbose", 0, 0, 1),
 	DEV_ITEM_STR("input_mode", "line"),
-	DEV_ITEM_INT("external_input_QL", 0, 0, 15),
-	DEV_ITEM_INT("external_input_ext_QL", 0, 0, 255),
 	DEV_ITEM_INT("extended_tlv", 0, 0, 1),
 	DEV_ITEM_INT("network_option", 1, 1, 2),
 	DEV_ITEM_INT("recover_time", 300, 10, 720),
@@ -165,12 +185,22 @@ struct config_item config_tab_synce[] = {
 	DEV_ITEM_STR("eec_locked_value", NULL),
 	DEV_ITEM_STR("eec_freerun_value", NULL),
 	DEV_ITEM_STR("eec_invalid_value", NULL),
+	DEV_ITEM_U64("clock_id", 0, 0, 0xffffffffffffffff),
+	DEV_ITEM_STR("module_name", NULL),
+	DEV_ITEM_INT("dnu_prio", 15, 15, 0xffff),
 	PORT_ITEM_STR("allowed_qls", NULL),
 	PORT_ITEM_STR("allowed_ext_qls", NULL),
 	PORT_ITEM_STR("recover_clock_enable_cmd", NULL),
 	PORT_ITEM_STR("recover_clock_disable_cmd", NULL),
+	PORT_ITEM_STR("board_label", NULL),
+	PORT_ITEM_STR("panel_label", NULL),
+	PORT_ITEM_STR("package_label", NULL),
 	PORT_ITEM_INT("tx_heartbeat_msec", 1000, 100, 3000),
 	PORT_ITEM_INT("rx_heartbeat_msec", 50, 10, 500),
+	EXT_ITEM_INT("input_QL", 0, 0, 15),
+	EXT_ITEM_INT("input_ext_QL", 0, 0, 255),
+	EXT_ITEM_STR("external_enable_cmd", NULL),
+	EXT_ITEM_STR("external_disable_cmd", NULL),
 };
 
 static struct interface *__config_create_interface(const char *name, struct config *cfg,
@@ -186,7 +216,9 @@ static struct config_item *config_section_item(struct config *cfg,
 {
 	char buf[CONFIG_LABEL_SIZE + MAX_IFNAME_SIZE];
 
-	snprintf(buf, sizeof(buf), "%s.%s", section, name);
+	if ((unsigned long)snprintf(buf, sizeof(buf), "%s.%s", section, name) >=
+	    sizeof(buf))
+		return NULL;
 	return hash_lookup(cfg->htab, buf);
 }
 
@@ -259,11 +291,14 @@ static enum parser_result parse_section_line(char *s, enum config_section *secti
 		char c;
 		if (s[1] == '<')
 			*section = DEVICE_SECTION;
+		else if (s[1] == '{')
+			*section = EXT_SECTION;
 		else
 			*section = PORT_SECTION;
 		/* Replace brackets with white space. */
 		while (0 != (c = *s)) {
-			if (c == '[' || c == ']' || c == '<' || c == '>')
+			if (c == '[' || c == ']' || c == '<' || c == '>' ||
+			    c == '{' || c == '}')
 				*s = ' ';
 			s++;
 		}
@@ -278,11 +313,12 @@ static enum parser_result parse_item(struct config *cfg,
 				     const char *option,
 				     const char *value)
 {
-	enum parser_result r;
 	struct config_item *cgi, *dst;
 	struct config_enum *cte;
-	double df;
-	int val;
+	enum parser_result r;
+	uint64_t u64 = 0;
+	double df = 0.0;
+	int val = 0;
 
 	r = parse_fault_interval(cfg, section, option, value);
 	if (r != NOT_PARSED)
@@ -300,6 +336,9 @@ static enum parser_result parse_item(struct config *cfg,
 	case CFG_TYPE_INT:
 		r = get_ranged_int(value, &val, cgi->min.i, cgi->max.i);
 		break;
+	case CFG_TYPE_U64:
+		r = get_ranged_u64(value, &u64, cgi->min.u64, cgi->max.u64);
+		break;
 	case CFG_TYPE_DOUBLE:
 		r = get_ranged_double(value, &df, cgi->min.d, cgi->max.d);
 		break;
@@ -322,6 +361,7 @@ static enum parser_result parse_item(struct config *cfg,
 
 	if (section) {
 		if (!(cgi->flags & CFG_ITEM_PORT) &&
+		    !(cgi->flags & CFG_ITEM_EXT) &&
 		    !(cgi->flags & CFG_ITEM_DEVICE)) {
 			return NOT_PARSED;
 		}
@@ -346,6 +386,9 @@ static enum parser_result parse_item(struct config *cfg,
 	case CFG_TYPE_ENUM:
 		dst->val.i = val;
 		break;
+	case CFG_TYPE_U64:
+		dst->val.u64 = u64;
+		break;
 	case CFG_TYPE_DOUBLE:
 		dst->val.d = df;
 		break;
@@ -455,8 +498,8 @@ int config_read(const char *name, struct config *cfg)
 	FILE *fp;
 	char buf[1024], *line, *c;
 	const char *option, *value;
-	struct interface *current_port = NULL;
 	struct interface *current_device = NULL;
+	struct interface *current_clk_src = NULL;
 	int line_num;
 
 	fp = 0 == strncmp(name, "-", 2) ? stdin : fopen(name, "r");
@@ -485,25 +528,30 @@ int config_read(const char *name, struct config *cfg)
 			*c-- = '\0';
 
 		if (parse_section_line(line, &current_section) == PARSED_OK) {
-			if (current_section == PORT_SECTION) {
-				char port[IF_NAMESIZE + 1];
-				if (1 != sscanf(line, " %16s", port)) {
-					fprintf(stderr, "could not parse port name on line %d\n",
+			if (current_section == PORT_SECTION ||
+			    current_section == EXT_SECTION) {
+				char clk_src[IF_NAMESIZE + 1];
+
+				if (sscanf(line, " %16s", clk_src) != 1) {
+					fprintf(stderr, "could not parse clk_src name on line %d\n",
 							line_num);
 					goto parse_error;
 				}
-				current_port = config_create_interface(port, cfg);
-				if (!current_port)
+				current_clk_src = config_create_interface(clk_src, cfg);
+				if (!current_clk_src)
 					goto parse_error;
 				if (current_device) {
-					interface_se_set_parent_dev(current_port,
+					interface_se_set_parent_dev(current_clk_src,
 						interface_name(current_device));
+					if (current_section == EXT_SECTION)
+						interface_section_set_external_source(
+							current_clk_src);
 				} else {
 					goto parse_error;
 				}
 			} else if (current_section == DEVICE_SECTION) {
-				/* clear port on new device found in config */
-				current_port = NULL;
+				/* clear clk_src on new device found in config */
+				current_clk_src = NULL;
 				char device[IF_NAMESIZE + 1];
 				if (1 != sscanf(line, " %16s", device)) {
 					fprintf(stderr, "could not parse device name on line %d\n",
@@ -525,14 +573,14 @@ int config_read(const char *name, struct config *cfg)
 		if (parse_setting_line(line, &option, &value)) {
 			fprintf(stderr, "could not parse line %d in %s section\n",
 				line_num, current_section == GLOBAL_SECTION ?
-				"global" : interface_name(current_port ?
-							  current_port : current_device));
+				"global" : interface_name(current_clk_src ?
+							  current_clk_src : current_device));
 			goto parse_error;
 		}
 
 		parser_res = parse_item(cfg, 0, current_section == GLOBAL_SECTION ?
-					NULL : interface_name(current_port ?
-							      current_port : current_device),
+					NULL : interface_name(current_clk_src ?
+							      current_clk_src : current_device),
 					option, value);
 		switch (parser_res) {
 		case PARSED_OK:
@@ -541,8 +589,8 @@ int config_read(const char *name, struct config *cfg)
 			fprintf(stderr, "unknown option %s at line %d in %s section\n",
 				option, line_num,
 				current_section == GLOBAL_SECTION ? "global" :
-				interface_name(current_port ?
-					       current_port : current_device));
+				interface_name(current_clk_src ?
+					       current_clk_src : current_device));
 			goto parse_error;
 		case BAD_VALUE:
 			fprintf(stderr, "%s is a bad value for option %s at line %d\n",
@@ -629,7 +677,11 @@ struct config *config_create()
 	for (i = 0; i < end; i++) {
 		ci = &ci_tab[i];
 		ci->flags |= CFG_ITEM_STATIC;
-		snprintf(buf, sizeof(buf), "global.%s", ci->label);
+		if ((unsigned long)snprintf(buf, sizeof(buf), "global.%s",
+					    ci->label) >= sizeof(buf)) {
+			fprintf(stderr, "option %s too long\n", ci->label);
+			goto fail;
+		}
 		if (hash_insert(cfg->htab, buf, ci)) {
 			fprintf(stderr, "duplicate item %s\n", ci->label);
 			goto fail;
@@ -691,16 +743,35 @@ int config_get_int(struct config *cfg, const char *section, const char *option)
 	switch (ci->type) {
 	case CFG_TYPE_DOUBLE:
 	case CFG_TYPE_STRING:
+	case CFG_TYPE_U64:
 		pr_err("bug: config option %s type mismatch!", option);
 		exit(-1);
 	case CFG_TYPE_INT:
 	case CFG_TYPE_ENUM:
 		break;
 	}
-	pr_debug("config item %s.%s is %d", section, option, ci->val.i);
+	pr_debug("config item %s.%s is %d (0x%x)", section, option, ci->val.i, ci->val.i);
 	return ci->val.i;
 }
 
+uint64_t
+config_get_u64(struct config *cfg, const char *section, const char *option)
+{
+	struct config_item *ci = config_find_item(cfg, section, option);
+
+	if (!ci) {
+		pr_err("bug: config option %s missing!", option);
+		exit(-1);
+	}
+	if (ci->type != CFG_TYPE_U64) {
+		pr_err("bug: config option %s type mismatch!", option);
+		exit(-1);
+	}
+	pr_debug("config item %s.%s is %lu (0x%lx)", section, option,
+		 ci->val.u64, ci->val.u64);
+	return ci->val.u64;
+}
+
 char *config_get_string(struct config *cfg, const char *section,
 			const char *option)
 {
@@ -768,6 +839,7 @@ int config_set_section_int(struct config *cfg, const char *section,
 	switch (cgi->type) {
 	case CFG_TYPE_DOUBLE:
 	case CFG_TYPE_STRING:
+	case CFG_TYPE_U64:
 		pr_err("bug: config option %s type mismatch!", option);
 		return -1;
 	case CFG_TYPE_INT:
